#include "game.h"

#include <vector>

#include <cstdlib>
#include <sys/time.h>
#include <GL/glut.h>

#include "vector3.h"

typedef struct bunny_desc {
    float size;
    float speed;
} BunnyDesc;

const float WORLD_SIZE          = 10.0f;
const float WALL_THICKNESS      = 0.1f;
const float WALL_HEIGHT         = 1.0f;

const float PLAYER_SIZE         = 0.2f;
const float PLAYER_SPEED        = 2.0f;

const float BUNNY_SIZE_BIG      = 1.0;
const float BUNNY_SPEED_BIG     = 0.7;

const float BUNNY_SIZE_MEDIUM   = 0.5;
const float BUNNY_SPEED_MEDIUM  = 0.9;

const float BUNNY_SIZE_SMALL    = 0.25;
const float BUNNY_SPEED_SMALL   = 0.7;

const float BULLET_SIZE         = 0.1f;
const float BULLET_SPEED        = 10.0f;

static const float GRAVITY      = 2.0f;

static const int NEW_BUNNIES_MIN = 3;
static const int NEW_BUNNIES_MAX = 5;

static const BunnyDesc bunnyDescriptions[] = {
    {BUNNY_SIZE_BIG,    BUNNY_SPEED_BIG},
    {BUNNY_SIZE_MEDIUM, BUNNY_SPEED_MEDIUM},
    {BUNNY_SIZE_SMALL,  BUNNY_SPEED_SMALL}
};
static const int BIG    = 0;
static const int MEDIUM = 1;
static const int SMALL  = 2;


Viewport view(800, 600);
Camera camera(Vector3(0.0, 0.5, 0.0), Vector3(0.0, 0.0, 1.0));
Mesh *bunnyMesh;
std::vector<Entity *> bunnies_big;
std::vector<Entity *> bunnies_medium;
std::vector<Entity *> bunnies_small;
std::vector<Movable *> bullets;

int moveForward = 0;
int moveBackward = 0;
int strafeLeft = 0;
int strafeRight = 0;


static void newBunnies(std::vector<Entity *> &bunnies, const BunnyDesc &desc, const Entity *bunny)
{
    int n = NEW_BUNNIES_MIN + rand() % (NEW_BUNNIES_MAX - NEW_BUNNIES_MIN);

    while (n--) {
        Entity *e = new Entity(*bunny);

        /* position */
        double r = 2.0 * M_PI * (double)rand() / RAND_MAX;
        double x = 2.0 * desc.size * cos(r);
        double z = 2.0 * desc.size * sin(r);
        e->translate(x, 0.0, z);

        /* orientation */
        r = 360.0 * (double)rand() / RAND_MAX;

        e->rotate(r, Vector3(0.0, 1.0, 0.0));
        Vector3 velocity = e->forward().unit() * desc.speed;
        e->setVelocity(velocity.x(), velocity.y(), velocity.z());
        bunnies.push_back(e);
    }
}

static void fixCamera(void)
{
    /* primitive collision detecion between camera and walls; clips the camera
     * to the world extents */
    Vector3 eye = camera.eye();
    if (eye.x() > WORLD_SIZE - PLAYER_SIZE)
        eye.setX(WORLD_SIZE - PLAYER_SIZE);
    if (eye.x() < -(WORLD_SIZE - PLAYER_SIZE))
        eye.setX(-(WORLD_SIZE - PLAYER_SIZE));
    if (eye.z() > WORLD_SIZE - PLAYER_SIZE)
        eye.setZ(WORLD_SIZE - PLAYER_SIZE);
    if (eye.z() < -(WORLD_SIZE - PLAYER_SIZE))
        eye.setZ(-(WORLD_SIZE - PLAYER_SIZE));
    camera.setEye(eye);
}

static void fixBunny(Entity *bunny, const BunnyDesc &desc)
{
    Vector3 position = bunny->location();
    Vector3 velocity = bunny->velocity();

    double px = position.x();
    double pz = position.z();
    double vx = velocity.x();
    double vz = velocity.z();

    /* collide with -X wall */
    if (px < -(WORLD_SIZE - desc.size)) {
        px -= 2 * (px - -(WORLD_SIZE - desc.size));
        vx = -vx;
    }

    /* collide with +X wall */
    if (px > +(WORLD_SIZE - desc.size)) {
        px -= 2 * (px - +(WORLD_SIZE - desc.size));
        vx = -vx;
    }

    /* collide with -Z wall */
    if (pz < -(WORLD_SIZE - desc.size)) {
        pz -= 2 * (pz - -(WORLD_SIZE - desc.size));
        vz = -vz;
    }

    /* collide with +Z wall */
    if (pz > +(WORLD_SIZE - desc.size)) {
        pz -= 2 * (pz - +(WORLD_SIZE - desc.size));
        vz = -vz;
    }

    bunny->setLocation(px, 0.0, pz);
    bunny->setVelocity(vx, 0.0, vz);
}

static Entity *testBullet(const Movable *bullet, std::vector<Entity *> &bunnies, const BunnyDesc &desc)
{
    std::vector<Entity *>::iterator itr;
    Entity *e;

    for (itr = bunnies.begin(); itr != bunnies.end(); itr++) {
        Vector3 diff = (*itr)->location() - bullet->location();
        if (diff.len() < (desc.size + BULLET_SIZE)) {
            /* hit bunny! */
            e = *itr;
            bunnies.erase(itr);
            return e;
        }
    }

    return NULL;
}

static void testBullets(void)
{
    std::vector<Movable *>::iterator itr_b;
    Entity *e;

    itr_b = bullets.begin();
    while (itr_b != bullets.end()) {
        if ((e = testBullet(*itr_b, bunnies_big, bunnyDescriptions[BIG]))) {
            newBunnies(bunnies_medium, bunnyDescriptions[MEDIUM], e);
            delete e;
            delete *itr_b;
            itr_b = bullets.erase(itr_b);
            continue;
        }

        if ((e = testBullet(*itr_b, bunnies_medium, bunnyDescriptions[MEDIUM]))) {
            newBunnies(bunnies_small, bunnyDescriptions[SMALL], e);
            delete e;
            delete *itr_b;
            itr_b = bullets.erase(itr_b);
            continue;
        }

        if ((e = testBullet(*itr_b, bunnies_small, bunnyDescriptions[SMALL]))) {
            delete e;
            delete *itr_b;
            itr_b = bullets.erase(itr_b);
            continue;
        }

        itr_b++;
    }
}

static void testCollision(const std::vector<Entity *> &bunnies, const BunnyDesc &desc)
{
    std::vector<Entity *>::const_iterator itr;

    for (itr = bunnies.begin(); itr != bunnies.end(); itr++) {
        Vector3 diff = (*itr)->location() - camera.eye();
        diff.setY(0.0); // ignore Z component

        if (diff.len() < (desc.size + PLAYER_SIZE)) {
            puts("you lose");
            exit(0);
        }
    }
}

static void calcGravity(Movable *obj, double timestep)
{
    Vector3 r = camera.eye() - obj->location();
    double rlen = r.len();
    Vector3 a = r * GRAVITY / (rlen * rlen * rlen) * (1.0 + 0.25 * rand() / RAND_MAX);
    Vector3 v = obj->velocity();

    obj->setVelocity(v + a*timestep);
}


void physics(void)
{
    std::vector<Entity *>::iterator itr;
    std::vector<Movable *>::iterator itr_b;

    static struct timeval prev_time = {0, 0};
    static struct timeval this_time = {0, 0};
    static struct timeval time_diff = {0, 0};

    if (!prev_time.tv_sec && !prev_time.tv_usec) {
    	/* calculate initial time */
        gettimeofday(&prev_time, NULL);
        return;
    }

    /* calculate time step */
    gettimeofday(&this_time, NULL);
    time_diff.tv_sec = this_time.tv_sec - prev_time.tv_sec;
    time_diff.tv_usec = this_time.tv_usec - prev_time.tv_usec;
    prev_time = this_time;

    double timestep = time_diff.tv_sec + time_diff.tv_usec / 1000000.0;

    /* move bunnies */
    for (itr = bunnies_big.begin(); itr != bunnies_big.end(); itr++) {
        (*itr)->doMovement(timestep);
        fixBunny(*itr, bunnyDescriptions[BIG]);
        calcGravity(*itr, timestep);
    }

    for (itr = bunnies_medium.begin(); itr != bunnies_medium.end(); itr++) {
        (*itr)->doMovement(timestep);
        fixBunny(*itr, bunnyDescriptions[MEDIUM]);
        calcGravity(*itr, timestep);
    }

    for (itr = bunnies_small.begin(); itr != bunnies_small.end(); itr++) {
        (*itr)->doMovement(timestep);
        fixBunny(*itr, bunnyDescriptions[SMALL]);
        calcGravity(*itr, timestep);
    }

    /* move bullets */
    itr_b = bullets.begin();
    while (itr_b != bullets.end()) {
        (*itr_b)->doMovement(timestep);

        if ((*itr_b)->location().len() > 2 * WORLD_SIZE) {
            delete *itr_b;
            itr_b = bullets.erase(itr_b);
            continue;
        }

        itr_b++;
    }

    /* calculate player movement */
    int moveZ = moveForward - moveBackward;
    int moveX = strafeRight - strafeLeft;

    if (moveZ) {
        /* move, but stay on ground */
        Vector3 v = camera.view();
        v.setY(0.0);
        v.normalize();
        camera.translate(v * moveZ * PLAYER_SPEED * timestep);
    }

    if (moveX)
        camera.strafe(moveX * PLAYER_SPEED * timestep);

    /* "collide" against walls */
    fixCamera();


    /* done moving objects; do collision-detection */
    testBullets();

    if (!bunnies_big.size() && !bunnies_medium.size() && !bunnies_small.size()) {
        puts("you win!");
        exit(0);
    }
    
    /* test collision of player against bunnies */
    testCollision(bunnies_big, bunnyDescriptions[BIG]);
    testCollision(bunnies_medium, bunnyDescriptions[MEDIUM]);
    testCollision(bunnies_small, bunnyDescriptions[SMALL]);

    glutPostRedisplay();
    glutWarpPointer(view.width()/2, view.height()/2);
}
